//
//  DocViewerController.m
//  HP41
//
//  Created by greg on Sun Aug 24 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

#import "DocViewerController.h"
#import "PDFView.h"
#import "PageTabView.h"
#import "PageTab.h"
#import "Documents.h"

static const float titleBarHeight = 20;

@implementation DocViewerController

- (NSString *) preferenceKey: (NSString *) name {
    return [NSString stringWithFormat: @"%@:%@", name, [[self window] title]];
}

- (id) preference: (NSString *) name {
    return [[NSUserDefaults standardUserDefaults] objectForKey: [self preferenceKey: name]];
}

- (void) setPreference: (NSString *) name to: (id) value {
    [[NSUserDefaults standardUserDefaults] setObject: value forKey: [self preferenceKey: name]];
}

- (int) intPreference: (NSString *) name withDefault: (int) dflt {
    id obj = [self preference: name];
    if (obj)
        return [obj intValue];
    else
        return dflt;
}

- (void) setIntPreference: (NSString *) name to: (int) value {
    [self setPreference: name to: [NSNumber numberWithInt: value]];
}

- (int) imageInterpolationPreference {
    return [self intPreference: @"ImageInterpolation" withDefault: NSImageInterpolationHigh];
}

- (void) setImageInterpolationPreference: (int) value {
    [self setIntPreference: @"ImageInterpolation" to: value];
}

- (float) tabViewWidth {
    return [self->leftTabView bounds].size.width;
}

- (float) extraControlsHeight {
    float h;
    if (self->extraControlsVisible)
        h = [self->extraControls frame].size.height;
    else
        h = 0;
    return h;
}

- (NSSize) windowSize {
    return [[[self window] contentView] frame].size;
}

- (NSSize) windowToPageSize: (NSSize) windowSize {
    NSSize pageSize;
    if (self->mode == onePageMode)
        pageSize.width = windowSize.width - [self tabViewWidth];
    else
        pageSize.width = windowSize.width / 2 - [self tabViewWidth];
    pageSize.height = windowSize.height /*- titleBarHeight*/ - [self extraControlsHeight];
    return pageSize;
}

- (NSSize) pageToWindowSize: (NSSize) pageSize {
    NSSize windowSize;
    if (self->mode == onePageMode)
        windowSize.width = pageSize.width + [self tabViewWidth];
    else
        windowSize.width = 2 * (pageSize.width + [self tabViewWidth]);
    windowSize.height = pageSize.height /*+ titleBarHeight*/ + [self extraControlsHeight];
    return windowSize;
}

- (void) layOutViews {
    NSRect r;
    float tabViewWidth = [self tabViewWidth];
    NSSize windowSize = [self windowSize]; //[[[self window] contentView] frame].size;
    NSSize pageSize = [self windowToPageSize: windowSize];
    float bottomOfPageViews = [self extraControlsHeight];
    
    r.origin.y = bottomOfPageViews;
    r.size.width = tabViewWidth;
    r.size.height = pageSize.height;
    r.origin.x = 0;
    [self->leftTabView setFrame: r];
    r.origin.x = windowSize.width - tabViewWidth;
    [self->rightTabView setFrame: r];
    
    r.origin.x = tabViewWidth;
    r.size = pageSize;
    [self->leftPageView setFrame: r];
    if (self->mode == onePageMode)
        r.origin.x = 0;
    else
        r.origin.x = NSMaxX(r);
    [self->rightPageView setFrame: r];
    
    r = [self->extraControls frame];
    r.origin.y = bottomOfPageViews - r.size.height;
    [self->extraControls setFrame: r];
}

- (id) initWithDesc: (struct DocumentDesc *) desc {
    [NSBundle loadNibNamed: @"DocViewer" owner: self];
    [[self window] setTitle: desc->title];
    [[self window] setFrameAutosaveName: desc->title];
    //self->mode = onePageMode;
    //self->mode = twoPageMode;
    self->mode = [self intPreference: @"PageMode" withDefault: onePageMode];
    //self->extraControlsVisible = YES;
    self->extraControlsVisible = [self intPreference: @"ShowControls" withDefault: NO];
    [self->leftPageView setWhichSide: leftSide];
    [self->rightPageView setWhichSide: rightSide];
    [self->leftTabView setWhichSide: leftSide];
    [self->rightTabView setWhichSide: rightSide];
    if (![self installPdfResource: desc->rsrcName])
        return nil;
    //[self layOutViews];
    self->currentPage = -1;
    [self goToPage: [self intPreference: @"Page" withDefault: 0]];
    return self;
}

- (void) dealloc {
    [self->pdf release];
    [self->tabs release];
    
    [super dealloc];
}

- (NSWindow *) window {
    return self->window;
}

- (BOOL) isShown {
    return [[self window] isVisible];
}

- (void) show: (BOOL) shown {
    if (shown)
        [[self window] makeKeyAndOrderFront: self];
    else
        [[self window] orderOut: self];
}

- (void) update {
    [[[self window] contentView] setNeedsDisplay: YES];
}

- (void) setWindowSize: (NSSize) newSize {
    [[self window] setContentSize: newSize];
    [self layOutViews];
    [self update];
};

- (NSSize) adjustPageSize: (NSSize) proposedSize {
    if (self->pdf) {
        NSRect pdfRect = [self->pdf bounds];
        float xscale = proposedSize.width / pdfRect.size.width;
        float yscale = proposedSize.height / pdfRect.size.height;
        float scale = xscale < yscale ? xscale : yscale;
        return NSMakeSize(scale * pdfRect.size.width, scale * pdfRect.size.height);
    }
    else
        return proposedSize;
}

- (NSSize) adjustWindowSize: (NSSize) proposedWindowSize {
    NSSize proposedPageSize = [self windowToPageSize: proposedWindowSize];
    NSSize newPageSize = [self adjustPageSize: proposedPageSize];
    NSSize newWindowSize = [self pageToWindowSize: newPageSize];
    return newWindowSize;
}

- (NSArray *) readTabsFromFile: (NSString *) filename {
    FILE *f = fopen([filename UTF8String], "r");
    if (!f)
        return nil;
    NSMutableArray *tabarray = [NSMutableArray array];
    for (;;) {
        int physPage;
        char labelChars[16];
        if (fscanf(f, "%d %15s", &physPage, labelChars) < 2)
            break;
        PageTab *tab = [[PageTab alloc] init];
        tab->physicalPageNumber = physPage;
        tab->label = [[NSString stringWithCString: labelChars] retain];
        [tabarray addObject: tab];
    }
    fclose(f);
    return tabarray;
}
    

- (NSArray *) buildDefaultTabs {
    NSMutableArray *tabarray = [NSMutableArray array];
    int numPages = [self numPages];
    int i;
    for (i = 0; i < numPages; i += 10) {
        PageTab *tab = [[PageTab alloc] init];
        tab->physicalPageNumber = i;
        if (i > 0)
            tab->label = [[NSString stringWithFormat: @"%d", i] retain];
        else
            tab->label = nil;
        [tabarray addObject: tab];
    }
    return tabarray;
}

- (void) installPdfFile: (NSString *) filename {
    NSData *data = [NSData dataWithContentsOfFile: filename];
    [self->pdf autorelease];
    self->pdf = [[NSPDFImageRep alloc] initWithData: data];
    self->currentPage = 0;
    //NSSize pdfSize = [self->pdf bounds].size;
    //[self setWindowSize: [self pageToWindowSize: pdfSize]];
    [self setWindowSize: [self adjustWindowSize: [self windowSize]]];
}

- (BOOL) installPdfResource: (NSString *) rsrcName {
    NSString *pdfFilename = [[NSBundle mainBundle] pathForResource: rsrcName ofType: @"pdf"];
    NSString *tabFilename = [[NSBundle mainBundle] pathForResource: rsrcName ofType: @"pagetabs"];
    if (pdfFilename == nil) {
        NSString *appDir = [[[NSBundle mainBundle] bundlePath] stringByDeletingLastPathComponent];
        pdfFilename = [appDir stringByAppendingFormat: @"/%@.pdf", rsrcName];
        //printf("pdfFilename = %s\n", [pdfFilename cString]); //***
        if (![[NSFileManager defaultManager] fileExistsAtPath: pdfFilename]) {
            //fprintf(stderr, "Resource %s not found\n", [rsrcName cString]); //***
            NSRunCriticalAlertPanel(@"Document Not Found", 
                @"Unable to find '%@.pdf'. You will need to download"
                @" this file and place it in the same folder as the X-41 application. See the"
                @" README file that came with X-41 for more information.",
                nil, nil, nil,
                rsrcName);
            return NO;
        }
    }
    //printf("Found resource: %s\n", [pdfFilename cString]); //***
    [self installPdfFile: pdfFilename];
    NSArray *tabarray = nil;
    if (tabFilename)
        tabarray = [self readTabsFromFile: tabFilename];
    if (!tabarray)
        tabarray = [self buildDefaultTabs];
    if (tabarray) {
        [self->tabs autorelease];
        self->tabs = [tabarray retain];
    }
    return YES;
}

- (NSPDFImageRep *) pdf {
    return self->pdf;
}

- (int) numPages {
    return [self->pdf pageCount];
}

- (int) currentPage {
    return self->currentPage;
}

- (NSArray *) tabs {
    return self->tabs;
}

- (int) currentPageForSide: (enum LeftOrRight) side {
    if (side == leftSide)
        // round down to left page
        return ((self->currentPage + 1) & ~1) - 1;
    else
        // round up to right page
        return ((self->currentPage + 1) | 1) - 1;
}

- (BOOL) currentPageIsOnSide: (enum LeftOrRight) side {
    if (side == leftSide)
        return (self->currentPage & 1) != 0;
    else
        return (self->currentPage & 1) == 0;
}

- (int) pageNumberForPageView: (PDFView *) view {
    enum LeftOrRight side = [view whichSide];
    if (self->mode == twoPageMode)
        return [self currentPageForSide: side];
    else if ([self currentPageIsOnSide: side])
        return self->currentPage;
    else
        return -1;
}

#if 0
- (int) pageNumberForPageView: (PDFView *) view {
    int p = self->currentPage;
    if (self->mode == onePageMode) {
        if ((view == self->leftPageView && (p & 1)) ||
            (view == self->rightPageView && !(p & 1)))
                return p;
        else
            return -1;
    }
    else {
        if (view == self->leftPageView)
            return p - 1;
        else
            return p;
    }
}
#endif

- (BOOL) tabViewIsEnabled: (PageTabView *) view {
    int p = self->currentPage;
    return self->mode != onePageMode
        || (view == self->leftTabView && (p & 1))
        || (view == self->rightTabView && !(p & 1));
}

- (BOOL) allTabsAreTogether {
    return self->mode == onePageMode;
}

- (void) goToPage: (int) newPage {
    int numPages = [self numPages];
    if (newPage < 0)
        newPage = 0;
    if (newPage > numPages)
        newPage = numPages;
    if (newPage != self->currentPage) {
        self->currentPage = newPage;
        [self setIntPreference: @"Page" to: newPage];
        [self->physicalPageField setIntValue: newPage];
        [self update];
    }
}

- (void) goInDirection: (int) d {
    if (self->mode == twoPageMode)
        d *= 2;
    [self goToPage: self->currentPage + d];
}

- (IBAction) goToFrontPage: (id) sender {
    [self goToPage: 0];
}

- (IBAction) goToBackPage: (id) sender {
    [self goToPage: [self numPages] - 1];
}

- (IBAction) goToPreviousPage: (id) sender {
    [self goInDirection: -1];
}

- (IBAction) goToNextPage: (id) sender {
    [self goInDirection: 1];
}

- (void) pageViewClicked: (PDFView *) view atPoint: (NSPoint) p {
    int d;
    if (self->mode == onePageMode) {
        float w = [view bounds].size.width;
        if (p.x < w / 2)
            d = -1;
        else
            d = +1;
    }
    else {
        if (view == self->leftPageView)
            d = -1;
        else
            d = +1;
    }
    [self goInDirection: d];
}

- (IBAction) physicalPageNoChanged: (id) sender {
    [self goToPage: [self->physicalPageField intValue]];
}

- (NSSize) windowWillResize: (NSWindow *) sender toSize: (NSSize) proposedFrameSize {
    return [self adjustWindowSize: proposedFrameSize];
}

- (void) windowDidResize: (NSNotification *) notification {
    [self layOutViews];
}

- (NSRect) windowWillUseStandardFrame: (NSWindow *) sender defaultFrame: (NSRect) defaultFrame {
    defaultFrame.origin.x = [[self window] frame].origin.x;
    defaultFrame.size = [self adjustWindowSize: defaultFrame.size];
    return defaultFrame;
}

- (void) updateWindowSize {
    NSSize pageSize = [self->leftPageView frame].size;
    NSSize windowSize = [self pageToWindowSize: pageSize];
    [self setWindowSize: windowSize];
}

- (void) setMode: (enum DocViewerMode) newMode {
    if (self->mode != newMode) {
        self->mode = newMode;
        [self setIntPreference: @"PageMode" to: newMode];
        [self updateWindowSize];
        [self update];
    }
}

- (BOOL) validateMenuItem: (id) item {
    SEL action = [item action];
    if (action == @selector(singlePageCmd:)) {
        [item setState: self->mode == onePageMode];
        return YES;
    }
    if (action == @selector(doublePageCmd:)) {
        [item setState: self->mode == twoPageMode];
        return YES;
    }
    if (action == @selector(showControlsCmd:)) {
        [item setState: self->extraControlsVisible];
        return YES;
    }
    if (action == @selector(noSmoothingCmd:)) {
        [item setState: [self imageInterpolationPreference] == NSImageInterpolationNone];
        return YES;
    }
    if (action == @selector(lightSmoothingCmd:)) {
        [item setState: [self imageInterpolationPreference] == NSImageInterpolationLow];
        return YES;
    }
    if (action == @selector(heavySmoothingCmd:)) {
        [item setState: [self imageInterpolationPreference] == NSImageInterpolationHigh];
        return YES;
    }
    return NO;
}

- (IBAction) singlePageCmd: (id) sender {
    [self setMode: onePageMode];
}

- (IBAction) doublePageCmd: (id) sender {
    [self setMode: twoPageMode];
}

- (IBAction) showControlsCmd: (id) sender {
    self->extraControlsVisible = !self->extraControlsVisible;
    [self setIntPreference: @"ShowControls" to: self->extraControlsVisible];
    [self updateWindowSize];
    [self update];
}

- (IBAction) noSmoothingCmd: (id) sender {
    [self setImageInterpolationPreference: NSImageInterpolationNone];
}

- (IBAction) lightSmoothingCmd: (id) sender {
    [self setImageInterpolationPreference: NSImageInterpolationLow];
}

- (IBAction) heavySmoothingCmd: (id) sender {
    [self setImageInterpolationPreference: NSImageInterpolationHigh];
}

@end
